<?php

namespace FiloBlu\Flow\Model\Indexer;

use Magento\Catalog\Model\Product;
use Magento\CatalogInventory\Model\Stock\Status;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Indexer\ActionInterface;
use Magento\Framework\Mview\ActionInterface as MviewActionInterface;
use Magento\Framework\ObjectManagerInterface;
use Magento\InventoryApi\Api\Data\SourceItemInterface;
use Magento\InventoryApi\Api\SourceItemRepositoryInterface;
use Magento\InventoryApi\Api\SourceItemsSaveInterface;
use Magento\InventoryConfigurationApi\Api\GetStockItemConfigurationInterface;
use Magento\InventoryConfigurationApi\Exception\SkuIsNotAssignedToStockException;
use Magento\InventoryReservationsApi\Model\GetReservationsQuantityInterface;
use Psr\Log\LoggerInterface;

class Stock implements ActionInterface, MviewActionInterface
{
    const INDEXER_ID = 'filoblu_flow_stock';
    protected $stockId;
    /**
     * @var SourceItemRepositoryInterface
     */
    private $sourceItemRepository;
    /**
     * @var SearchCriteriaBuilder
     */
    private $searchCriteriaBuilder;
    /**
     * @var SourceItemsSaveInterface
     */
    private $sourceItemsSave;
    /**
     * @var ResourceConnection
     */
    private $resourceConnection;
    /**
     * @var Product
     */
    private $product;
    /**
     * @var LoggerInterface
     */
    private $logger;
    /**
     * @var GetReservationsQuantityInterface
     */
    private $getReservationsQuantity;
    /**
     * @var GetStockItemConfigurationInterface
     */
    private $getStockItemConfiguration;


    /**
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param ObjectManagerInterface $objectManager
     * @param ResourceConnection $resourceConnection
     * @param Product $product
     * @param LoggerInterface $logger
     * @param GetReservationsQuantityInterface $getReservationsQuantity
     * @param GetStockItemConfigurationInterface $getStockItemConfiguration
     */
    public function __construct(
        SearchCriteriaBuilder  $searchCriteriaBuilder,
        ObjectManagerInterface $objectManager,
        ResourceConnection     $resourceConnection,
        Product                $product,
        LoggerInterface        $logger,
        GetReservationsQuantityInterface $getReservationsQuantity,
        GetStockItemConfigurationInterface $getStockItemConfiguration
    )
    {
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
        if (interface_exists(SourceItemsSaveInterface::class)) {
            $this->sourceItemsSave = $objectManager->get(SourceItemsSaveInterface::class);
        }
        if (interface_exists(SourceItemRepositoryInterface::class)) {
            $this->sourceItemRepository = $objectManager->get(SourceItemRepositoryInterface::class);
        }

        $this->resourceConnection = $resourceConnection;
        $this->product = $product;
        $this->logger = $logger;
        $this->getReservationsQuantity = $getReservationsQuantity;
        $this->getStockItemConfiguration = $getStockItemConfiguration;
    }


    public function executeFull()
    {
        $this->execute(null);
    }

    public function execute($skus)
    {
        try {
            if (!$this->sourceItemsSave || !$this->sourceItemRepository) {
                return;
            }

            $this->searchCriteriaBuilder
                ->addFilter('source_code', 'default');
            if ($skus) {
                $this->searchCriteriaBuilder->addFilter('sku', $skus, 'in');
            }

            $result = $this->sourceItemRepository->getList(
                $this->searchCriteriaBuilder->create()
            );

            if (!$result->getTotalCount()) {
                return;
            }
            $connection = $this->resourceConnection->getConnection();
            $tableName = $connection->getTableName('cataloginventory_stock_status');

            $sourceItems = $result->getItems();
            $sourceItemToUpdate=[];

            foreach ($sourceItems as $sourceItem){
                $stockItemConfiguration = $this->getStockItemConfiguration->execute($sourceItem->getSku(), 1);
                // We don't need to Manage Stock
                if (!$stockItemConfiguration->isManageStock() || $stockItemConfiguration->getBackorders()) {
                    continue;
                }
                $sourceItemToUpdate[]=$sourceItem;
            }


            $columnsSql = $this->buildColumnsSqlPart([
                Status::KEY_PRODUCT_ID,
                Status::KEY_WEBSITE_ID,
                Status::KEY_STOCK_ID,
                Status::KEY_QTY,
                Status::KEY_STOCK_STATUS
            ]);

            $valuesSql = $this->buildValuesSqlPart($sourceItemToUpdate);

            $onDuplicateSql = $this->buildOnDuplicateSqlPart([
                Status::KEY_QTY,
                Status::KEY_STOCK_STATUS
            ]);

            $bind = $this->getSqlBindData($sourceItemToUpdate);

            $insertSql = sprintf(
                'INSERT INTO `%s` (%s) VALUES %s %s',
                $tableName,
                $columnsSql,
                $valuesSql,
                $onDuplicateSql
            );

            $connection->query($insertSql, $bind);

        } catch (\Exception $e) {
            $this->logger->critical("FiloBlu_Flow_Stock:" . $e->getMessage());
        } catch (\Throwable $t) {
            $this->logger->critical("FiloBlu_Flow_Stock:" . $t->getMessage());
        }
    }


    /**
     * @param array $columns
     * @return string
     */
    private function buildColumnsSqlPart(array $columns): string
    {
        $connection = $this->resourceConnection->getConnection();
        $processedColumns = array_map([$connection, 'quoteIdentifier'], $columns);
        $sql = implode(', ', $processedColumns);
        return $sql;
    }

    /**
     * @param array $sourceItems
     * @return string
     */
    private function buildValuesSqlPart(array $sourceItems): string
    {
        $sql = rtrim(str_repeat('(?, ?, ?, ?, ?), ', count($sourceItems)), ', ');
        return $sql;
    }

    /**
     * @param array $fields
     * @return string
     */
    private function buildOnDuplicateSqlPart(array $fields): string
    {
        $connection = $this->resourceConnection->getConnection();
        $processedFields = [];
        foreach ($fields as $field) {
            $processedFields[] = sprintf('%1$s = VALUES(%1$s)', $connection->quoteIdentifier($field));
        }
        $sql = 'ON DUPLICATE KEY UPDATE ' . implode(', ', $processedFields);
        return $sql;
    }

    /**
     * @param array $sourceItems
     * @return array
     * @throws LocalizedException
     * @throws SkuIsNotAssignedToStockException
     */
    private function getSqlBindData(array $sourceItems): array
    {
        /*
         *  Status::KEY_PRODUCT_ID,
            Status::KEY_WEBSITE_ID,
            Status::KEY_STOCK_ID,
            Status::KEY_QTY,
            Status::KEY_STOCK_STATUS
         */
        $bind = [];
        foreach ($sourceItems as $sourceItem) {
            $reservations = $this->getReservationsQuantity->execute($sourceItem->getSku(), 1);
            $minimumQuantity = $this->getStockItemConfiguration->execute($sourceItem->getSku(), 1)->getMinQty();
            $saleableQty= ($sourceItem->getQuantity() + $reservations - $minimumQuantity >= 0 ) ? $sourceItem->getQuantity() + $reservations - $minimumQuantity : 0;
            $realStockStatus = ($sourceItem->getQuantity() + $reservations - $minimumQuantity > 0) ? SourceItemInterface::STATUS_IN_STOCK : SourceItemInterface::STATUS_OUT_OF_STOCK;
            $bind[] = $this->product->getIdBySku($sourceItem->getSku());
            $bind[] = 0;
            $bind[] = 1;
            $bind[] = $saleableQty;
            $bind[] = $realStockStatus;
        }
        return $bind;
    }

    public function executeList(array $skus)
    {
        $this->execute($skus);
    }

    public function executeRow($sku)
    {
        $this->execute([$sku]);
    }
}
